"""The available reporting formats."""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from __future__ import annotations

from enum import Enum


class ReportFormat(Enum):
    """The available reporting formats."""

    TABLE = "table"
    HTML = "html"
    CYCLONE_DX = "cyclone_dx"
    GITLAB = "gitlab"



"""Provides vulnerability reporting."""
# This file is part of hoppr-security-commons
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin CorporationØ
from __future__ import annotations

import json
import uuid

from collections import defaultdict
from pathlib import Path
from typing import Any

import jinja2
import typer

from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    Advisory,
    Affect,
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
    Rating,
    ScoreMethod,
    Severity,
    Vulnerability,
)
from packageurl import PackageURL
from tabulate import tabulate

from hoppr_security_commons.reporting.models import ReportFormat


class Reporting:
    """Generates reports in multiple formats from a list of vulnerabilities."""

    output_path: Path
    base_name: str

    def __init__(self, output_path: Path, base_name: str):
        self.output_path = output_path
        self.base_name = base_name

    def generate_vulnerability_reports(
        self,
        formats: list[ReportFormat],
        vulnerabilities: dict[str, list[Vulnerability] | None],
        bom: Bom_1_4 | None = None,
    ):
        """Generates various vulnerability reports based on specified formats."""
        flattened_vulnerabilties = self.__add_purl_as_bom_ref_and_flatten(vulnerabilities)

        if ReportFormat.CYCLONE_DX in formats and bom is not None:
            self.output_path.mkdir(parents=True, exist_ok=True)
            self.add_vulnerabilities_to_bom(bom, flattened_vulnerabilties)

            (self.output_path / f"{Path(self.base_name).name}-enhanced.json").write_text(
                bom.json(indent=2), encoding="utf-8"
            )

        if ReportFormat.GITLAB in formats:
            self.generate_gitlab_vulnerability_report(flattened_vulnerabilties)

        if ReportFormat.HTML in formats:
            self.__generate_html_report(flattened_vulnerabilties)

        if ReportFormat.TABLE in formats:
            findings = self.__get_fields_from_vulnerabilities(flattened_vulnerabilties)
            typer.echo(tabulate(findings, headers=["type", "name", "version", "id", "severity", "found by"]))

    @staticmethod
    def add_vulnerabilities_to_bom(bom: Bom_1_4, vulnerabilities: list[Vulnerability]) -> Bom_1_4:
        """Adds the vulnerabilities found by various scanners to cyclone 1.4+ compliant BOM."""
        if isinstance(bom, Bom_1_4):
            bom.vulnerabilities = vulnerabilities
        else:
            typer.echo("Cannot add vulnerabilities to a bom earlier than cyclone version 1.4")

        return bom

    @staticmethod
    def link_vulnerabilities_to_bom(vulnerabilities: list[Vulnerability]) -> Bom_1_4:
        """Creates/Adds the vulnerabilities found by various scanners to cyclone 1.4+ compliant VEX BOM."""
        vex_bom = Bom_1_4(bomFormat="CycloneDX", specVersion="1.4", version=1)
        vex_bom.serialNumber = f"urn:uuid:{uuid.uuid4()}"
        vex_bom.vulnerabilities = vulnerabilities

        return vex_bom

    def __add_purl_as_bom_ref_and_flatten(
        self, vulnerabilities: dict[str, list[Vulnerability] | None]
    ) -> list[Vulnerability]:
        flattened_vulnerabilities: list[Vulnerability] = []

        for purl in vulnerabilities:
            for vuln in vulnerabilities[purl] or []:
                vuln.affects = vuln.affects or []
                vuln.affects.append(Affect(ref=purl))
                flattened_vulnerabilities.append(vuln)

        flattened_vulnerabilities.sort(key=self.get_score, reverse=True)
        return flattened_vulnerabilities

    def get_score(self, vuln_to_score: Vulnerability) -> float:
        """Return best score of specified Vulnerability."""
        best_rating = self.__get_best_rating(vuln_to_score.ratings)

        if best_rating is None or best_rating.score is None:
            return 0.0

        return best_rating.score

    def __get_fields_from_vulnerabilities(self, vulnerabilities: list[Vulnerability]) -> list[list[str | None]]:
        findings = []

        def get_fields(vuln: Vulnerability) -> list[str | None]:
            tools = [f"{tool.vendor} {tool.name}" for tool in vuln.tools or []]
            severity = self.__get_severity(vuln.ratings)

            if severity == "critical":
                severity = typer.style(severity, fg=typer.colors.RED)
            elif severity == "high":
                severity = typer.style(severity, fg=typer.colors.BRIGHT_YELLOW)

            if not vulnerability.affects:
                return []

            purl = PackageURL.from_string(vulnerability.affects[0].ref)

            return [purl.type, purl.name, purl.version, vuln.id, severity, " | ".join(tools)]

        for vulnerability in vulnerabilities:
            findings.append(get_fields(vulnerability))

        return findings

    def __copy_assets(self):
        assets_dir = Path(__file__).parent / "templates" / "assets"
        assets = ["vulnerabilities.css"]

        output_path = self.output_path / "assets"
        output_path.mkdir(exist_ok=True, parents=True)

        for asset in assets:
            template_data = (assets_dir / asset).read_text(encoding="utf-8")
            (output_path / asset).write_text(template_data)

    def __generate_html_report(self, combined_list: list[Vulnerability]):
        self.output_path.mkdir(parents=True, exist_ok=True)
        output_path = self.output_path / f"{Path(self.base_name).name}-vulnerabilities.html"

        env = jinja2.Environment(loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"))
        env.filters["severity"] = self.__get_severity
        template = env.get_template("vulnerabilities.html")

        self.__copy_assets()

        severity_classes = {
            "critical": "bg-red-100 rounded-lg py-5 px-6 mb-4 text-base text-red-700 mb-3",
            "high": "bg-yellow-100 rounded-lg py-5 px-6 mb-4 text-base text-yellow-700 mb-3",
            "medium": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "info": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "low": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "unknown": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "none": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
        }

        result = template.render(
            {"findings": combined_list, "severity_classes": severity_classes, "base_name": self.base_name},
        )

        output_path.write_text(result, encoding="utf-8")

        self.__generate_vuln_detail_reports(combined_list, severity_classes)

    def __generate_vuln_detail_reports(self, vulnerabilities: list[Vulnerability], severity_classes: dict[str, str]):
        output_path = self.output_path / f"{Path(self.base_name).name}-details"
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"))
        env.filters["featured_link"] = self.__get_featured_link
        template = env.get_template("vulnerability_details.html")

        for vuln in vulnerabilities:
            if not vuln.affects:
                continue

            purl = PackageURL.from_string(vuln.affects[0].ref)
            result = template.render(
                {
                    "type": purl.type,
                    "namespace": purl.namespace,
                    "name": purl.name,
                    "version": purl.version,
                    "severity_classes": severity_classes,
                    "vulnerability": vuln,
                    "purl": purl.to_string(),
                    "base_name": self.base_name,
                },
            )

            (output_path / f"{vuln.id}.html").write_text(result, encoding="utf-8")

    def generate_gitlab_vulnerability_report(self, vulnerabilities: list[Vulnerability]):
        """Renders the vulnerabilities report for gitlab."""
        # Note the JSON schema for this report can be found at
        # https://gitlab.com/gitlab-org/security-products/security-report-schemas/-/blob/master/dist/dependency-scanning-report-format.json
        report: dict[str, Any] = {
            "version": "14.1.2",
            "vulnerabilities": [],
            "remediations": [],
            "dependency_files": [],
        }

        dependencies_by_format: dict[str, list[Any]] = defaultdict(list, {})

        purls = list({vuln.affects[0].ref for vuln in vulnerabilities if vuln.affects})

        self.output_path.mkdir(parents=True, exist_ok=True)
        output_path = self.output_path / "gl-dependency-scanning-report.json"

        for purl_str in purls:
            purl = PackageURL.from_string(purl_str)
            dependencies_by_format[purl.type].append({"package": {"name": purl.name}, "version": purl.version})

        for vuln in vulnerabilities:
            report["vulnerabilities"].append(self.__generate_gitlab_row(vuln))

        for repo_format in dependencies_by_format:
            report["dependency_files"].append(
                {
                    "package_manager": repo_format,
                    "path": "cyclonedx.bom",
                    "dependencies": dependencies_by_format[repo_format],
                },
            )

        output_path.write_text(json.dumps(report, indent=4, sort_keys=True, default=str), encoding="utf-8")

    @staticmethod
    def __get_featured_link(advisories: list[Advisory] | None) -> str | None:
        if advisories is not None:
            for adv in advisories:
                url = "" if adv.url is None else adv.url
                if "https://snyk.io/" in url:
                    return url
        return None

    def __get_severity(self, ratings: list[Rating] | None) -> str:
        best_rating = self.__get_best_rating(ratings)
        if best_rating is None or best_rating.severity is None:
            return "none"

        return Severity(best_rating.severity).value

    @staticmethod
    def __get_best_rating(ratings: list[Rating] | None) -> Rating | None:
        default_rating = ratings[0] if ratings else None

        methods = [ScoreMethod(rating.method).value if rating.method else "none" for rating in ratings or []]

        preferred_method = None
        if "CVSSv31" in methods:
            preferred_method = "CVSSv31"
        elif "CVSSv3" in methods:
            preferred_method = "CVSSv3"
        elif "CVSSv2" in methods:
            preferred_method = "CVSSv2"

        return next(
            filter(lambda rating: rating.method and (str(rating.method) == preferred_method), ratings or []),
            default_rating,
        )

    def __generate_gitlab_row(self, vuln: Vulnerability) -> dict[str, Any] | None:
        """Generates a report row."""
        # Ensure `affects` and `tools` are non-empty lists
        if not vuln.affects or not vuln.tools:
            return None

        purl = PackageURL.from_string(vuln.affects[0].ref)
        severity = self.__get_severity(vuln.ratings).title()

        return {
            "category": "dependency_scanning",
            "name": vuln.description,
            "description": vuln.description,
            "cve": vuln.id,
            "severity": severity if severity != "none" else "Info",
            "confidence": "Unknown",
            "identifiers": [{"type": "cve", "name": vuln.id, "value": vuln.id}],
            "scanner": {"id": vuln.tools[0].name, "name": vuln.tools[0].name},
            "location": {
                "file": "cyclonedx.bom",
                "dependency": {"package": {"name": purl.name}, "version": purl.version},
            },
        }



"""This provides a set of utility functions for working with vulnerabilities and boms."""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from __future__ import annotations

import json
import os
import stat
import tempfile
import uuid

from pathlib import Path
from typing import TYPE_CHECKING

import requests
import typer

from hoppr_cyclonedx_models.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr_cyclonedx_models.cyclonedx_1_4 import (
    Advisory,
    Component,
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
    Reference,
    VulnerabilitySource,
)


if TYPE_CHECKING:
    from packageurl import PackageURL


def convert_xml_to_json(file_path: Path) -> Path:
    """Function to convert a xml file to json format."""
    typer.echo("xml format detected, attempt to convert with cyclonedx tools")

    # Default to the path specified in the hoppr-cop docker file or define the local filename to save data
    docker_image_path = Path("/usr/local/bin/cyclone-dx")
    cyclone_dx_path = docker_image_path if docker_image_path.exists() else Path(tempfile.gettempdir()) / "cyclonedx"
    if not cyclone_dx_path.exists():
        typer.echo("cyclonedx tools not found Attempting to download")
        if url := os.getenv("CYCLONE_INSTALL_URL"):
            # Make http request for remote file datae
            data = requests.get(url, timeout=60)

            # Save file data to local copy
            cyclone_dx_path.write_bytes(data.content)
        else:
            msg = typer.style(
                "In order to support xml boms, you must set 'CYCLONE_INSTALL_URL' to "
                "the correct release of cyclone-dx cli. https://github.com/CycloneDX/cyclonedx-cli/releases",
                fg=typer.colors.RED,
            )
            typer.echo(msg)
            raise typer.Exit(code=1)

    cyclone_dx_path.chmod(stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

    os.system(
        f"{cyclone_dx_path.absolute()} convert --input-file {file_path} \
            --output-file {tempfile.gettempdir()}/{file_path.name}.json --output-format json",
    )
    return Path(f"{tempfile.gettempdir()}/{file_path.name}.json")


def parse_sbom(sbom_file: Path) -> Bom_1_4 | Bom_1_3:
    """Parses a Software Bill of Materials File."""
    if sbom_file.is_file() and sbom_file.exists():
        typer.echo(f"processing {sbom_file}")

        if sbom_file.suffix == ".xml":
            sbom_file = convert_xml_to_json(sbom_file)

        sbom_json_object = json.loads(sbom_file.read_text(encoding="utf-8"))
        sbom = create_sbom_object(sbom_json_object, str(sbom_file))
    else:
        typer.secho(f"{sbom_file!s} is not a file", fg=typer.colors.RED)
        raise typer.Exit

    return sbom


def parse_sbom_json_string(sbom_file_json: str, sbom_info: str) -> Bom_1_4 | Bom_1_3:
    """Parses a Software Bill of Materials JSON String."""
    sbom_json_object = json.loads(sbom_file_json)
    return create_sbom_object(sbom_json_object, sbom_info)


def create_sbom_object(sbom_json_object: dict, sbom_info: str) -> Bom_1_4 | Bom_1_3:
    """Creates a Software Bill of Materials Object."""
    spec_version = sbom_json_object.get("specVersion", "")
    sbom_json_object.pop("$schema", None)

    sbom: Bom_1_4 | Bom_1_3
    if spec_version == "1.4":
        sbom = Bom_1_4(**sbom_json_object)
    elif spec_version == "1.3":
        sbom = Bom_1_3(**sbom_json_object)
    else:
        typer.secho(f"{sbom_info} is an unknown spec version ({spec_version})")
        raise typer.Exit

    return sbom


def get_vulnerability_source(vulnerability_id: str) -> VulnerabilitySource | None:
    """Generate the source for a vulnerability based on a given ID."""
    if vulnerability_id.startswith("CVE-"):
        return VulnerabilitySource(name="NVD", url=f"https://nvd.nist.gov/vuln/detail/{vulnerability_id}")
    if vulnerability_id.startswith("GHSA"):
        return VulnerabilitySource(name="Github Advisories", url=f"https://github.com/advisories/{vulnerability_id}")
    if vulnerability_id.startswith("GMS"):
        return VulnerabilitySource(name="Github Advisories", url=f"https://github.com/advisories/{vulnerability_id}")
    if vulnerability_id.startswith("sonatype"):
        return VulnerabilitySource(
            name="OSS Index",
            url=f"https://ossindex.sonatype.org/vulnerability/{vulnerability_id}",
        )

    return None


def get_advisories_from_urls(urls: list[str]) -> list[Advisory]:
    """Generates a list of advisories for the given set of urls."""
    urls = list(set(urls))
    return [Advisory(url=x) for x in urls]


def get_references_from_ids(ids: list[str], primary_id: str) -> list[Reference]:
    """Builds a list of Reference objects to the given vulnerability IDs."""
    references = []

    for ident in list(set(ids)):
        if ident != primary_id and (source := get_vulnerability_source(ident)):
            references.append(Reference(id=ident, source=source))

    return references


def build_bom_dict_from_purls(purls: list[PackageURL]) -> dict:
    """Create SBOM dictionary from PackageURL list."""
    sbom = Bom_1_4(specVersion="1.4", serialNumber=uuid.uuid1().urn)
    sbom.components = []  # assign explicitly for type checkers

    for purl in purls:
        component = Component(
            type="library",
            name=purl.name,
            version=purl.version,
            purl=purl.to_string(),
            group=purl.namespace,
            bom_ref=purl.to_string(),
            description="test",
            author="test",
            externalReferences=[],
        )

        sbom.components.append(component)

    return sbom.dict()


def build_bom_from_purls(purls: list[PackageURL]) -> Bom_1_4:
    """Creates a skeleton BOM from a list of purls."""
    return Bom_1_4(**build_bom_dict_from_purls(purls))



"""Defines the common interface for a vulnerability scanner."""
# This file is part of hoppr-cop
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Copyright (c) 2022 Lockheed Martin Corporation
from __future__ import annotations

import distutils.spawn
import os

from typing import TYPE_CHECKING, ClassVar

import typer

from packageurl import PackageURL


if TYPE_CHECKING:
    from hoppr_cyclonedx_models.cyclonedx_1_3 import (
        CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
    )
    from hoppr_cyclonedx_models.cyclonedx_1_4 import (
        CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
        Vulnerability,
    )


class VulnerabilityMeta(type):
    """a meta class to define interface expectations of a vulnerability scanner."""

    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return cls.__subclasscheck__(type(instance))

    @classmethod
    def __subclasscheck__(cls, subclass: type) -> bool:
        return (
            hasattr(subclass, "get_vulnerabilities_by_purl")
            and callable(subclass.get_vulnerabilities_by_purl)
            and hasattr(subclass, "get_vulnerabilities_by_sbom")
            and callable(subclass.get_vulnerabilities_by_sbom)
            and hasattr(subclass, "should_activate")
            and callable(subclass.should_activate)
        )


class VulnerabilitySuper(metaclass=VulnerabilityMeta):
    """A super class that defines the expected interface for a vulnerability scanner."""

    required_tools_on_path: ClassVar[list[str]] = []
    required_environment_variables: ClassVar[list[str]] = []
    offline_mode_supported = False
    offline_mode = False

    def get_vulnerabilities_by_purl(self, purls: list[PackageURL]) -> dict[str, list[Vulnerability] | None]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        return {}

    def get_vulnerabilities_by_sbom(self, bom: Bom_1_4 | Bom_1_3) -> dict[str, list[Vulnerability] | None]:
        """Accepts a cyclone dx compatible BOM and returns a list of vulnerabilities.

        Return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        purls = [
            PackageURL.from_string(component.purl)
            for component in bom.components or []
            if component.purl is not None and component.purl != ""
        ]

        return self.get_vulnerabilities_by_purl(purls)

    def should_activate(self) -> bool:
        """Checks if the vulnerability scanner should activate based on it's requirements."""
        activate = True
        for tool in self.required_tools_on_path:
            if not distutils.spawn.find_executable(tool):
                activate = False
                typer.echo(f"{type(self).__name__} is not activated because {tool} is not on the PATH")
        for env in self.required_environment_variables:
            if env not in os.environ:
                activate = False
                typer.echo(
                    f"{type(self).__name__} is not activated because required environment variable {env} is not set",
                )

        if self.offline_mode and not self.offline_mode_supported:
            activate = False
            typer.echo(f"{type(self).__name__} is not activated because offline mode is not supported by this scanner")

        return activate
